/*
 * Miya.com Inc.
 * Copyright (c) 2004-2018 All Rights Reserved.
 */
package com.miya.draw.util;

import com.alibaba.fastjson.support.spring.FastJsonHttpMessageConverter;
import com.miya.draw.config.exception.BusinessException;
import com.miya.draw.enums.ErrorEnum;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.MapUtils;
import org.apache.http.client.HttpClient;
import org.apache.http.impl.client.DefaultConnectionKeepAliveStrategy;
import org.apache.http.impl.client.HttpClientBuilder;
import org.apache.http.impl.client.HttpClients;
import org.apache.http.impl.conn.PoolingHttpClientConnectionManager;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpHeaders;
import org.springframework.http.MediaType;
import org.springframework.http.client.HttpComponentsClientHttpRequestFactory;
import org.springframework.http.converter.ByteArrayHttpMessageConverter;
import org.springframework.http.converter.FormHttpMessageConverter;
import org.springframework.http.converter.HttpMessageConverter;
import org.springframework.http.converter.StringHttpMessageConverter;
import org.springframework.stereotype.Component;
import org.springframework.util.LinkedMultiValueMap;
import org.springframework.util.MultiValueMap;
import org.springframework.web.client.DefaultResponseErrorHandler;
import org.springframework.web.client.RestTemplate;

import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.TimeUnit;

/**
 * rest 请求工具类
 *
 * @author ${zhuoming}
 * @author liushuangzeng
 * @version $Id: RestTemplateUtils.java, v 0.1 2018年07月27日 11:29 ${zhuoming} Exp $
 */
@Slf4j
@Component
public class RestTemplateUtil {

    private static RestTemplate restTemplate;


    private static final int CONN_TIME_OUT = 3000;

    private static final int SOCKET_TIME_OUT = 5000;

    private static final int MAX_TOTAL = 500;

    private static final int MAX_PER_ROUTE = 200;

    private static final int RETRY_COUNT = 2;

    private static final int CONN_REQUEST_TIME_OUT = 200;


    static {

        // 长连接保持30秒
        PoolingHttpClientConnectionManager pollingConnectionManager = new PoolingHttpClientConnectionManager(30, TimeUnit.SECONDS);

        // 总连接数
        pollingConnectionManager.setMaxTotal(MAX_TOTAL);

        // 同路由的并发数
        pollingConnectionManager.setDefaultMaxPerRoute(MAX_PER_ROUTE);
        HttpClientBuilder httpClientBuilder = HttpClients.custom();
        httpClientBuilder.setConnectionManager(pollingConnectionManager);
        // 保持长连接配置，需要在头添加Keep-Alive
        httpClientBuilder.setKeepAliveStrategy(new DefaultConnectionKeepAliveStrategy());
        HttpClient httpClient = httpClientBuilder.build();
        // httpClient连接配置，底层是配置RequestConfig
        HttpComponentsClientHttpRequestFactory clientHttpRequestFactory = new HttpComponentsClientHttpRequestFactory(httpClient);
        // 连接超时
        clientHttpRequestFactory.setConnectTimeout(CONN_TIME_OUT);
        // 数据读取超时时间，即SocketTimeout
        clientHttpRequestFactory.setReadTimeout(SOCKET_TIME_OUT);
        // 连接不够用的等待时间，不宜过长，必须设置，比如连接不够用时，时间过长将是灾难性的
        clientHttpRequestFactory.setConnectionRequestTimeout(CONN_REQUEST_TIME_OUT);


        // 添加内容转换器
        List<HttpMessageConverter<?>> messageConverters = new ArrayList<>();
        messageConverters.add(new StringHttpMessageConverter(Charset.forName("UTF-8")));
        messageConverters.add(new ByteArrayHttpMessageConverter());
        messageConverters.add(new FormHttpMessageConverter());

        //设置fastjson支持的格式
        FastJsonHttpMessageConverter fastJsonHttpMessageConverter = new FastJsonHttpMessageConverter();
        List<MediaType> supportedMediaTypes = new ArrayList<>();
        supportedMediaTypes.add(MediaType.APPLICATION_JSON);
        supportedMediaTypes.add(MediaType.APPLICATION_ATOM_XML);
        supportedMediaTypes.add(MediaType.APPLICATION_FORM_URLENCODED);
        supportedMediaTypes.add(MediaType.APPLICATION_OCTET_STREAM);
        supportedMediaTypes.add(MediaType.APPLICATION_PDF);
        supportedMediaTypes.add(MediaType.APPLICATION_XHTML_XML);
        supportedMediaTypes.add(MediaType.APPLICATION_XML);
        supportedMediaTypes.add(MediaType.IMAGE_GIF);
        supportedMediaTypes.add(MediaType.IMAGE_JPEG);
        supportedMediaTypes.add(MediaType.IMAGE_PNG);
        supportedMediaTypes.add(MediaType.TEXT_HTML);
        supportedMediaTypes.add(MediaType.TEXT_MARKDOWN);
        supportedMediaTypes.add(MediaType.TEXT_PLAIN);
        supportedMediaTypes.add(MediaType.TEXT_XML);
        fastJsonHttpMessageConverter.setSupportedMediaTypes(supportedMediaTypes);

        // 用FastJson替代原生的jackson转换器
        messageConverters.add(fastJsonHttpMessageConverter);

        restTemplate = new RestTemplate(messageConverters);
        restTemplate.setRequestFactory(clientHttpRequestFactory);
        restTemplate.setErrorHandler(new DefaultResponseErrorHandler());

        log.info("restTemplate initialized successfully.");
    }

    /**
     * post请求，body为实体对象的json形式
     *
     * @param url
     * @param jsonObject 能序列化的实体对象参数
     * @return
     */
    public static String postWithJson(String url, Object jsonObject) {

        if (null == jsonObject) {
            return postWithEmpty(url);
        }
        return postForObject(url, jsonObject);
    }

    public static String postWithJsonWithRetry(String url, Object jsonObject) {
        StringBuilder builder = new StringBuilder();

        RetryUtils.retry(() -> builder.append(postWithJson(url, jsonObject)));

        return builder.toString();
    }

    /**
     * 带请求头的重试
     *
     * @param url        url
     * @param jsonObject jsonObject
     * @param headers    请求头
     * @return String
     */
    public static String postWithJsonWithRetry(String url, Object jsonObject, Map<String, ?> headers) {
        StringBuilder builder = new StringBuilder();

        RetryUtils.retry(() -> builder.append(postWithJson(url, jsonObject, headers)));

        return builder.toString();
    }

    /**
     * @param url
     * @param jsonObject
     * @param headers
     * @return
     */
    public static String postWithJson(String url, Object jsonObject, Map<String, ?> headers) {
        if (null == jsonObject) {
            return postWithEmpty(url);
        }
        //拿到header信息
        HttpHeaders requestHeaders = new HttpHeaders();
        requestHeaders.setContentType(MediaType.APPLICATION_JSON);
        headers.forEach((x, y) -> requestHeaders.add(x, y.toString()));
        // 构建entity
        HttpEntity entity = new HttpEntity(jsonObject, requestHeaders);
        log.info("http url={} entity={}",url, entity);
        Object result = restTemplate.postForObject(url, entity, Object.class);
        log.info("http request={} result={}", jsonObject, result);
        return result.toString();

    }

    /**
     * @param url
     * @return
     */
    public static String get(String url) {
        StringBuilder builder = new StringBuilder();

        RetryUtils.retry(() -> builder.append(getForObject(url, new HashMap<>())));

        return builder.toString();
    }

    public static String getWithRetry(String url) {
        return getForObject(url, new HashMap<>());
    }


    /**
     * 以表单的形式提交参数
     *
     * @param url        地址
     * @param formParams 参数
     * @return
     */
    public static String postWithForm(String url, Map<String, Object> formParams) {
        return postWithForm(url, formParams, null);
    }

    /**
     * 以表单的形式提交参数
     *
     * @param url        地址
     * @param formParams 参数
     * @param headers    请求头
     * @return
     */
    public static String postWithForm(String url, Map<String, Object> formParams, Map<String, String> headers) {

        HttpHeaders httpHeaders = new HttpHeaders();
        httpHeaders.setContentType(MediaType.APPLICATION_FORM_URLENCODED);

        // 设置请求头
        if (MapUtils.isNotEmpty(headers)) {
            headers.forEach(httpHeaders::add);
        }

        // 组装参数
        MultiValueMap<String, Object> params = new LinkedMultiValueMap<>();

        if (MapUtils.isNotEmpty(formParams)) {
            formParams.forEach(params::add);
        }

        // 构建entity
        HttpEntity<MultiValueMap<String, Object>> entity = new HttpEntity<>(params, httpHeaders);

        try {
            return restTemplate.postForObject(url, entity, String.class);
        } catch (Exception e) {
            log.error("POST请求出错，url=" + url + ", errorMsg=" + e.getMessage() + ", params=" + formParams, e);
            throw new BusinessException(ErrorEnum.HTTP_REQ_EXP);
        }

    }

    /**
     * @param url
     * @param paramMap key为参数名，value为参数值
     * @return
     */
    public static String getWithMap(String url, Map<String, Object> paramMap) {
        return getForObject(url, paramMap);
    }

    /**
     * @param url
     * @return
     */
    private static String postWithEmpty(String url) {
        return postForObject(url, HttpEntity.EMPTY);
    }

    /**
     * 实际调用restTemplate的get请求
     *
     * @param url
     * @param paramMap
     * @return
     */
    private static String getForObject(String url, Map<String, Object> paramMap) {
        try {
            return restTemplate.getForObject(url, String.class, paramMap);
        } catch (Exception e) {
            log.error("GET请求出错，url=" + url + ", errorMsg=" + e.getMessage(), e);
            throw new BusinessException(ErrorEnum.HTTP_REQ_EXP);
        }
    }

    /**
     * 实际调用restTemplate的post请求
     *
     * @param url
     * @param object
     * @return
     */
    private static String postForObject(String url, Object object) {
        try {
            return restTemplate.postForObject(url, object, String.class);
        } catch (Exception e) {
            log.error("POST请求出错，url=" + url + ", errorMsg=" + e.getMessage(), e);
            throw new BusinessException(ErrorEnum.HTTP_REQ_EXP);
        }


    }
}