package com.miya.draw.util;

import com.miya.draw.config.exception.BusinessException;
import com.miya.draw.enums.ErrorEnum;
import lombok.extern.slf4j.Slf4j;

/**
 * 重试工具
 *
 * @Author wensh.zhu
 * @Date 2019/4/30 16:37
 */
@Slf4j
public class RetryUtils {

    /**
     * 默认重试次数:3
     */
    private static final int DEFAULT_RETRY_TIMES = 3;

    /**
     * 默认重试间隔：1秒
     */
    private static final long DEFAULT_RETRY_INTERVAL = 1000L;

    /**
     * 重试3次，每次间隔1秒
     *
     * @param func
     */
    public static void retry(Func func) {
        retry(func, DEFAULT_RETRY_TIMES, DEFAULT_RETRY_INTERVAL);
    }

    /**
     * 重试
     *
     * @param func       重试操作
     * @param retryTimes 重试次数
     * @param intervals  重试间隔，单位：毫秒
     */
    public static void retry(Func func, int retryTimes, long intervals) {
        if (func == null) {
            throw new NullPointerException("Func can not be null");
        }

        if (retryTimes <= 0 || intervals <= 0L) {
            throw new NumberFormatException("retryTimes and intervals must be greater than 0");
        }

        Exception exception = null;
        boolean success = false;
        for (int i = 0; i < retryTimes; i++) {
            try {
                func.call();
                success = true;
                break;
            } catch (Exception e) {
                exception = e;
                log.error("重试报错:{}", e.getMessage());
            }

            try {
                Thread.sleep(intervals);
            } catch (InterruptedException e) {
                log.error("重试休眠被打断", e);
            }
        }

        if (!success) {
            throw new BusinessException(ErrorEnum.RETRY_EXP, "重试失败", exception);
        }
    }


    public interface Func {

        void call();
    }
}
